//  (C) Copyright John Maddock 2005.
//  Use, modification and distribution are subject to the
//  Boost Software License, Version 1.0. (See accompanying file
//  LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)

//
// Forwarding header for Borland C++:
//
#if !defined(BOOST_TR1_TYPE_TRAITS_H_INCLUDED)
#  define BOOST_TR1_TYPE_TRAITS_H_INCLUDED
#  include <type_traits.>
#endif

